import unittest
import zope.component.testing

from zope.interface import implementer

from zope.component import adapter
from zope.component import provideHandler
from zope.component.event import objectEventNotify

from zope.event import notify

from zope.lifecycleevent import ObjectModifiedEvent
from zope.lifecycleevent import ObjectAddedEvent
from zope.lifecycleevent import ObjectRemovedEvent
from zope.lifecycleevent import ObjectMovedEvent

from z3c.caching.interfaces import IPurgeEvent
from z3c.caching.interfaces import IPurgeable

from z3c.caching.purge import purgeOnModified
from z3c.caching.purge import purgeOnMovedOrRemoved


class Handler(object):

    def __init__(self):
        self.invocations = []

    @adapter(IPurgeEvent)
    def handler(self, event):
        self.invocations.append(event)


class FauxContainer(dict):
    pass


class FauxContext(object):

    def __init__(self, parent=None, name=None):
        self.__parent__ = parent
        self.__name__ = name


@implementer(IPurgeable)
class FauxMarkedContext(object):

    def __init__(self, parent=None, name=None):
        self.__parent__ = parent
        self.__name__ = name


class TestPurgeRedispatch(unittest.TestCase):

    def setUp(self):
        self.handler = Handler()
        provideHandler(self.handler.handler)

        provideHandler(objectEventNotify)
        provideHandler(purgeOnModified)
        provideHandler(purgeOnMovedOrRemoved)

    def tearDown(self):
        zope.component.testing.tearDown()

    def test_no_marker(self):
        context = FauxContext(FauxContainer(), 'new')

        notify(ObjectModifiedEvent(context))
        notify(ObjectAddedEvent(context))
        notify(ObjectRemovedEvent(context))

        self.assertEqual(0, len(self.handler.invocations))

    def test_modified(self):
        context = FauxMarkedContext()

        notify(ObjectModifiedEvent(context))

        self.assertEqual(1, len(self.handler.invocations))
        self.assertEqual(context, self.handler.invocations[0].object)

    def test_added(self):
        context = FauxMarkedContext(FauxContainer(), 'new')

        notify(ObjectAddedEvent(context, context.__parent__, 'new'))

        self.assertEqual(0, len(self.handler.invocations))

    def test_moved(self):
        context = FauxMarkedContext(FauxContainer(), 'new')

        notify(ObjectMovedEvent(context, FauxContainer(), 'old',
                                context.__parent__, 'new'))

        self.assertEqual(1, len(self.handler.invocations))
        self.assertEqual(context, self.handler.invocations[0].object)

    def test_renamed(self):
        context = FauxMarkedContext(FauxContainer(), 'new')

        notify(ObjectMovedEvent(context,
                                context.__parent__, 'old',
                                context.__parent__, 'new'))

        self.assertEqual(1, len(self.handler.invocations))
        self.assertEqual(context, self.handler.invocations[0].object)

    def test_removed(self):
        context = FauxMarkedContext(FauxContainer(), 'new')

        notify(ObjectRemovedEvent(context, context.__parent__, 'new'))

        self.assertEqual(1, len(self.handler.invocations))
        self.assertEqual(context, self.handler.invocations[0].object)


def test_suite():
    return unittest.defaultTestLoader.loadTestsFromName(__name__)



from unittest import TestCase
import zope.component.testing

from zope.interface import implementer
from zope.component import adapter, provideAdapter
from zope.browser.interfaces import IView

from z3c.caching.interfaces import ILastModified
from z3c.caching.lastmodified import viewDelegateLastModified


class TestLastModified(TestCase):

    def setUp(self):
        provideAdapter(viewDelegateLastModified)

    def tearDown(self):
        zope.component.testing.tearDown()

    def test_no_adapter(self):

        @implementer(IView)
        class DummyView(object):

            def __init__(self, context, request):
                self.context = context
                self.request = request

        class DummyContext(object):
            pass

        class DummyRequest(dict):
            pass

        context = DummyContext()
        request = DummyRequest()

        view = DummyView(context, request)

        lastModified = ILastModified(view, None)
        self.assertIsNone(lastModified)

    def test_with_adapter(self):

        @implementer(IView)
        class DummyView(object):

            def __init__(self, context, request):
                self.context = context
                self.request = request

        class DummyContext(object):
            pass

        class DummyRequest(dict):
            pass

        @implementer(ILastModified)
        @adapter(DummyContext)
        class DummyLastModified(object):

            def __init__(self, context):
                self.context = context
                self.request = request

        provideAdapter(DummyLastModified)

        context = DummyContext()
        request = DummyRequest()

        view = DummyView(context, request)

        lastModified = ILastModified(view)
        self.assertIsInstance(lastModified, DummyLastModified)

        self.assertEqual(context, lastModified.context)
        self.assertEqual(request, lastModified.request)


def test_suite():
    import unittest
    return unittest.defaultTestLoader.loadTestsFromName(__name__)



from zope.interface import Interface
from zope.interface.interfaces import IObjectEvent
from zope import schema

class IRulesetRegistry(Interface):

    def register(obj, rule):
        """Mark objects that are implementers of `obj` to use the caching
        rule `rule`. The name should be a dotted name, consisting only of
        upper or lowercase letters, digits, and/or periods.
        """

    def unregister(obj):
        """Remove any prior rule registration attached to obj in this
        registry. N.B. registries are hierarchical, a parent may still
        provide rules.
        """

    def clear():
        """Remove all rule registrations in this registry.
        """

    def lookup(obj):
        """Return the id of the rule associated with `obj`.  If no rule has
        been registered `None` is returned.
        """

    def __getitem__(obj):
        """Convenience spelling for `lookup(obj)`.
        """

    def declareType(name, type, description):
        """Declare a new ruleset type. This will put a new `IRulesetType`
        into the list of objects returned by `enumerate`. The name should be
        a dotted name, consisting only of upper or lowercase letters, digits,
        and/or periods.
        """

    def enumerateTypes():
        """Return a sequence of all unique registered rule set types, as
        ``IRuleSetType`` objects.
        """

    explicit = schema.Bool(
            title=u"Explicit mode",
            description=u"If true, ruleset types must be declared before being used.",
            required=True,
            default=False
        )

class IRulesetType(Interface):
    """A ruleset type. The name can be used in a <cache:ruleset /> directive.
    The title and description are used for UI support.
    """

    name        = schema.DottedName(title=u"Ruleset name")
    title       = schema.TextLine(title=u"Title")
    description = schema.TextLine(title=u"Description", required=False)

class ILastModified(Interface):
    """An abstraction to help obtain a last-modified date for a published
    resource.

    Should be registered as an unnamed adapter from a published object
    (e.g. a view).
    """

    def __call__():
        """Return the last-modified date, as a Python datetime object.

        The datetime returned must be timezone aware and should normally be
        in the local timezone.

        May return None if the last modified date cannot be determined.
        """

class IPurgeEvent(IObjectEvent):
    """Event which can be fired to purge a particular object.

    This event is not fired anywhere in this package. Instead, higher level
    frameworks are expected to fire this event when an object may need to be
    purged.

    It is safe to fire the event multiple times for the same object. A given
    object will only be purged once.
    """

class IPurgeable(Interface):
    """Marker interface for content which should be purged when modified or
    removed.

    Event handlers are registered for ``IObjectModifiedEvent`` and
    ``IObjectRemovedEvent`` for contexts providing this interface. These are
    automatically purged.
    """

class IPurgePaths(Interface):
    """Return paths to send as PURGE requests for a given object.

    The purging hook will look up named adapters from the objects sent to
    the purge queue (usually by an IPurgeEvent being fired) to this interface.
    The name is not significant, but is used to allow multiple implementations
    whilst still permitting per-type overrides. The names should therefore
    normally be unique, prefixed with the dotted name of the package to which
    they belong.
    """

    def getRelativePaths():
        """Return a list of paths that should be purged. The paths should be
        relative to the virtual hosting root, i.e. they should start with a
        '/'.

        These paths will be rewritten to incorporate virtual hosting if
        necessary.
        """

    def getAbsolutePaths():
        """Return a list of paths that should be purged. The paths should be
        relative to the domain root, i.e. they should start with a '/'.

        These paths will *not* be rewritten to incorporate virtual hosting.
        """



from zope.interface import implementer
from zope.component import adapter
from zope.event import notify

from zope.lifecycleevent.interfaces import IObjectModifiedEvent
from zope.lifecycleevent.interfaces import IObjectMovedEvent

from z3c.caching.interfaces import IPurgeable
from z3c.caching.interfaces import IPurgeEvent

@implementer(IPurgeEvent)
class Purge(object):
    """Event implementation.

    To queue a purge for a given object, you can do::

        from plone.cachepurging import Purge
        from zope.event import notify

        notify(Purge(someobject))

    The actual URL(s) to purge are looked up via any relevant IPurgeURLs
    adapters.
    """

    def __init__(self, object):
        self.object = object


@adapter(IPurgeable, IObjectModifiedEvent)
def purgeOnModified(object, event):
    notify(Purge(object))

@adapter(IPurgeable, IObjectMovedEvent)
def purgeOnMovedOrRemoved(object, event):
    # Don't purge when added
    if event.oldName is not None and event.oldParent is not None:
        notify(Purge(object))



from z3c.caching.registry import getGlobalRulesetRegistry
from zope.configuration.fields import GlobalObject
from zope.interface import Interface
from zope.schema import TextLine

ORDER = 1000001


class IRuleset(Interface):
    for_ = GlobalObject(
            title=u"Object to be configured",
            description=u"The object for which the cache ruleset is to be defined",
            default=None,
            required=True)

    ruleset = TextLine(
            title=u"ruleset",
            description=u"The id of the cache ruleset to use",
            default=None,
            required=True)


def rulesetType(_context, name, title, description=u""):
    # The order is 'late' so that we know getGlobalRulesetRegistry() will
    # return something real. The default order of a configuration action is
    # determined by a counter, so in systems with a large number of components
    # registered, a big number is called for to ensure that we occur after
    # z3c.caching has been configured.
    _context.action(
            discriminator=("declareCacheRuleSetType", name),
            callable=declareType,
            args=(name, title, description,),
            order=ORDER)


def ruleset(_context, for_, ruleset):
    # We need to ensure this has a higher order than rulesetType() always
    _context.action(
            discriminator=("registerCacheRule", for_),
            callable=register,
            args=(for_, ruleset,),
            order=ORDER + 1)


# Handlers
def declareType(name, title, description):
    getGlobalRulesetRegistry().declareType(name, title, description)


def register(for_, ruleset):
    getGlobalRulesetRegistry().register(for_, ruleset)



from setuptools import setup, find_packages


version = '2.2'


def read(name):
    """Read a file in cwd."""
    with open(name) as f:
        return f.read()

setup(
    name="z3c.caching",
    version=version,
    description="Caching infrastructure for web apps",
    long_description="\n\n".join([
        read("README.rst"),
        read("CHANGES.rst")]),
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Zope Public License",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    keywords="zope caching",
    author="Wichert Akkerman",
    author_email="zope-dev@zope.org",
    url="https://github.com/zopefoundation/z3c.caching",
    license="ZPL",
    namespace_packages=["z3c"],
    packages=find_packages("src", exclude=["ez_setup"]),
    package_dir={"": "src"},
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        "setuptools",
        "zope.interface>=3.8.0",
        "zope.component",
        "zope.event",
        "zope.lifecycleevent",
        "zope.browser",
    ],
    extras_require={
        "zcml": ("zope.configuration", )
    },
    tests_require="nose >=0.10.0b1",
    test_suite="nose.collector",
)



from unittest import TestCase
import warnings
from zope.interface import Interface
from zope.interface import implementer

from zope.component import provideAdapter
from zope.component import provideUtility

from z3c.caching.registry import RulesetRegistry
from z3c.caching.registry import getGlobalRulesetRegistry

import zope.component.testing


class ITestView(Interface):
    pass


class IMoreSpecificTestView(ITestView):
    pass


@implementer(ITestView)
class TestView(object):
    pass


@implementer(IMoreSpecificTestView)
class OtherTestView(object):
    pass


class IDummy(Interface):
    pass


class TestRulesetRegistry(TestCase):

    def setUp(self):
        provideAdapter(RulesetRegistry)
        self.registry = getGlobalRulesetRegistry()

    def tearDown(self):
        self.registry.clear()
        zope.component.testing.tearDown()

    def test_no_ruleset_returned_if_unregistered(self):
        self.assertIsNone(self.registry[None])

    def test_ruleset_for_class(self):
        self.registry.register(TestView, "frop")
        i = TestView()
        self.assertEqual(self.registry[i], "frop")

    def test_ruleset_for_interface(self):
        self.registry.register(ITestView, "frop")
        i = TestView()
        self.assertEqual(self.registry[i], "frop")

    def test_most_specific_interface_wins(self):
        self.registry.register(ITestView, "frop")
        self.registry.register(IMoreSpecificTestView, "fribble")
        i = OtherTestView()
        self.assertEqual(self.registry[i], "fribble")

    def test_direct_lookup_for_interface_works(self):
        self.registry.register(ITestView, "frop")
        self.assertEqual(self.registry.directLookup(ITestView), "frop")

    def test_direct_lookup_for_daughter_fails(self):
        self.registry.register(ITestView, "frop")
        self.assertIsNone(self.registry.directLookup(IMoreSpecificTestView))

    def test_registration_on_class_ignores_any_interface_relationship(self):
        self.registry.register(TestView, "frop")

        i = OtherTestView()
        self.assertEqual(self.registry[i], None)

        i = TestView()
        self.assertEqual(self.registry[i], "frop")

    def test_registration_on_class_wins_over_interface_registration(self):
        self.registry.register(ITestView, "frop")
        self.registry.register(TestView, "fribble")

        i = TestView()
        self.assertEqual(self.registry[i], "fribble")

    def test_ruleset_registered_twice(self):
        self.registry.register(ITestView, "frop")

        # Hide the warning generated, that's for users, not tests.
        warnings.simplefilter("ignore")
        self.registry.register(ITestView, "fribble")
        warnings.simplefilter("default")

        i = TestView()
        self.assertEqual(self.registry[i], "frop")

    def test_unregistering_ruleset_removes_ruleset(self):
        self.registry.register(TestView, "frop")
        self.registry.unregister(TestView)
        self.assertIsNone(self.registry[TestView])

    def test_unregistering_nonexistant_ruleset_doesnt_error(self):
        self.assertIsNone(self.registry[TestView])
        self.registry.unregister(TestView)
        self.assertIsNone(self.registry[TestView])

    def test_clearing_registry_removes_rulesets(self):
        self.registry.register(ITestView, "frop")

        self.registry.clear()

        i = TestView()
        self.assertIsNone(self.registry[i])

    def test_clearing_ignores_other_utilities(self):

        @implementer(IDummy)
        class DummyUtility(object):
            pass

        provideUtility(DummyUtility())

        self.registry.register(ITestView, "frop")

        self.registry.clear()

        i = TestView()
        self.assertIsNone(self.registry[i])

    def test_clearing_registry_removes_types(self):
        self.registry.declareType("rule1", u"Rule 1", u"First rule")
        self.registry.declareType("rule2", u"Rule 2", u"Second rule")

        self.registry.register(ITestView, "frop")

        self.registry.clear()

        i = TestView()
        self.assertIsNone(self.registry[i])
        self.assertEqual(0, len(list(self.registry.enumerateTypes())))

    def test_declareType_overrides(self):
        self.registry.declareType("rule1", u"Rule 1", u"First rule")
        self.registry.declareType("rule2", u"Rule 2", u"Second rule")
        self.registry.declareType("rule1", u"Rule One", u"Rule uno")

        rules = list(self.registry.enumerateTypes())
        rules.sort(key=lambda x: x.name)

        self.assertEqual(2, len(rules))
        self.assertEqual("rule1", rules[0].name)
        self.assertEqual(u"Rule One", rules[0].title)
        self.assertEqual(u"Rule uno", rules[0].description)
        self.assertEqual("rule2", rules[1].name)
        self.assertEqual(u"Rule 2", rules[1].title)
        self.assertEqual(u"Second rule", rules[1].description)

    def test_enumerateTypes(self):
        self.registry.declareType("rule1", u"Rule 1", u"First rule")
        self.registry.declareType("rule2", u"Rule 2", u"Second rule")

        rules = list(self.registry.enumerateTypes())
        rules.sort(key=lambda x: x.title)

        self.assertEqual(2, len(rules))
        self.assertEqual("rule1", rules[0].name)
        self.assertEqual(u"Rule 1", rules[0].title)
        self.assertEqual(u"First rule", rules[0].description)
        self.assertEqual("rule2", rules[1].name)
        self.assertEqual(u"Rule 2", rules[1].title)
        self.assertEqual(u"Second rule", rules[1].description)

    def test_enumerate_empty(self):
        self.assertEqual(set([]), set(self.registry.enumerateTypes()))

    def test_set_explicit_mode(self):
        self.registry.explicit = True

        with self.assertRaises(LookupError):
            self.registry.register(TestView, "rule1")
        self.assertIsNone(self.registry.lookup(TestView()))

        self.registry.declareType("rule1", u"Rule 1", u"First rule")
        self.registry.register(TestView, "rule1")

        self.assertEqual("rule1", self.registry.lookup(TestView()))

    def test_disable_explicit_mode(self):
        self.registry.explicit = True

        with self.assertRaises(LookupError):
            self.registry.register(TestView, "rule1")
        self.assertIsNone(self.registry.lookup(TestView()))

        self.registry.explicit = False

        self.registry.register(TestView, "rule1")
        self.assertEqual("rule1", self.registry.lookup(TestView()))


class TestConvenienceAPI(TestCase):

    def setUp(self):
        provideAdapter(RulesetRegistry)
        self.registry = getGlobalRulesetRegistry()

    def tearDown(self):
        self.registry.clear()
        zope.component.testing.tearDown()

    def test_no_ruleset_returned_if_unregistered(self):
        from z3c.caching.registry import lookup
        self.assertIsNone(lookup(None))

    def test_ruleset_for_class(self):
        from z3c.caching.registry import register, lookup
        register(TestView, "frop")
        i = TestView()
        self.assertEqual(lookup(i), "frop")

    def test_ruleset_for_interface(self):
        from z3c.caching.registry import register, lookup
        register(ITestView, "frop")
        i = TestView()
        self.assertEqual(lookup(i), "frop")

    def test_most_specific_interface_wins(self):
        from z3c.caching.registry import register, lookup
        register(ITestView, "frop")
        register(IMoreSpecificTestView, "fribble")
        i = OtherTestView()
        self.assertEqual(lookup(i), "fribble")

    def test_unregistering_ruleset_removes_ruleset(self):
        from z3c.caching.registry import register, unregister, lookup
        register(TestView, "frop")
        unregister(TestView)
        self.assertIsNone(lookup(TestView))

    def test_unregistering_nonexistant_ruleset_doesnt_error(self):
        from z3c.caching.registry import unregister, lookup
        self.assertIsNone(lookup(TestView))
        unregister(TestView)
        self.assertIsNone(lookup(TestView))

    def test_declareType_enumerateTypes(self):
        from z3c.caching.registry import declareType, enumerateTypes
        declareType("rule1", u"Rule 1", u"Rule one")

        rules = list(enumerateTypes())
        rules.sort(key=lambda x: x.name)

        self.assertEqual(1, len(rules))
        self.assertEqual("rule1", rules[0].name)
        self.assertEqual(u"Rule 1", rules[0].title)
        self.assertEqual(u"Rule one", rules[0].description)

    def test_set_explicit_mode(self):
        from z3c.caching.registry import setExplicitMode

        self.assertFalse(self.registry.explicit)
        setExplicitMode()
        self.assertTrue(self.registry.explicit)
        setExplicitMode(False)
        self.assertFalse(self.registry.explicit)
        setExplicitMode(True)
        self.assertTrue(self.registry.explicit)


class TestNotSetUp(TestCase):

    def test_getGlobalRulesetRegistry(self):
        from z3c.caching.registry import getGlobalRulesetRegistry
        self.assertIsNone(getGlobalRulesetRegistry())

    def test_register(self):
        from z3c.caching.registry import register
        with self.assertRaises(LookupError):
            register(TestView, 'testrule')

    def test_unregister(self):
        from z3c.caching.registry import unregister
        with self.assertRaises(LookupError):
            unregister(TestView)

    def test_lookup(self):
        from z3c.caching.registry import lookup
        self.assertIsNone(lookup(TestView))

    def test_enumerateTypes(self):
        from z3c.caching.registry import enumerateTypes
        with self.assertRaises(LookupError):
            enumerateTypes()

    def test_declareType(self):
        from z3c.caching.registry import declareType
        with self.assertRaises(LookupError):
            declareType('name', 'title', 'description')

    def test_setExplicitMode(self):
        from z3c.caching.registry import setExplicitMode
        with self.assertRaises(LookupError):
            setExplicitMode()


def test_suite():
    import unittest
    return unittest.defaultTestLoader.loadTestsFromName(__name__)
